import pytest
from pytest_mock import MockerFixture

from artifex.core.decorators.logging import _serialize_value


@pytest.mark.unit
def test_serialize_value_with_none():
    """
    Test that _serialize_value returns None when given None.
    """
    result = _serialize_value(None)
    assert result is None


@pytest.mark.unit
def test_serialize_value_with_short_string():
    """
    Test that _serialize_value returns the string as-is when it's below max_length.
    """
    test_string = "Hello, World!"
    result = _serialize_value(test_string)
    assert result == test_string


@pytest.mark.unit
def test_serialize_value_with_long_string():
    """
    Test that _serialize_value truncates strings that exceed max_length.
    """
    test_string = "a" * 1500
    result = _serialize_value(test_string, max_length=1000)
    
    assert result.startswith("a" * 1000)
    assert "... (truncated, total length: 1500)" in result
    assert len(result) < len(test_string)


@pytest.mark.unit
def test_serialize_value_with_empty_list():
    """
    Test that _serialize_value returns an empty list when given an empty list.
    """
    result = _serialize_value([])
    assert result == []


@pytest.mark.unit
def test_serialize_value_with_small_list():
    """
    Test that _serialize_value returns a serialized list with all items when list is small.
    """
    test_list = [1, 2, 3, "hello", None]
    result = _serialize_value(test_list)
    
    assert isinstance(result, list)
    assert len(result) == 5
    assert result == ["1", "2", "3", "hello", None]


@pytest.mark.unit
def test_serialize_value_with_large_list():
    """
    Test that _serialize_value returns a summary for lists with more than 10 items.
    """
    test_list = list(range(20))
    result = _serialize_value(test_list)
    
    assert isinstance(result, dict)
    assert result["type"] == "list"
    assert result["count"] == 20
    assert "sample" in result
    assert len(result["sample"]) == 3
    assert result["sample"] == ["0", "1", "2"]


@pytest.mark.unit
def test_serialize_value_with_nested_list():
    """
    Test that _serialize_value recursively serializes nested lists.
    """
    test_list = [[1, 2], [3, 4, 5]]
    result = _serialize_value(test_list)
    
    assert isinstance(result, list)
    assert result == [["1", "2"], ["3", "4", "5"]]


@pytest.mark.unit
def test_serialize_value_with_empty_dict():
    """
    Test that _serialize_value returns an empty dict when given an empty dict.
    """
    result = _serialize_value({})
    assert result == {}


@pytest.mark.unit
def test_serialize_value_with_simple_dict():
    """
    Test that _serialize_value correctly serializes simple dictionaries.
    """
    test_dict = {"key1": "value1", "key2": "42", "key3": None}
    result = _serialize_value(test_dict)
    
    assert isinstance(result, dict)
    assert result == test_dict


@pytest.mark.unit
def test_serialize_value_with_nested_dict():
    """
    Test that _serialize_value recursively serializes nested dictionaries.
    """
    test_dict = {
        "outer": {
            "inner": {
                "deep": "value"
            }
        }
    }
    result = _serialize_value(test_dict)
    
    assert isinstance(result, dict)
    assert result["outer"]["inner"]["deep"] == "value"


@pytest.mark.unit
def test_serialize_value_with_dict_containing_list():
    """
    Test that _serialize_value handles mixed structures with dicts and lists.
    """
    test_dict = {
        "numbers": [1, 2, 3],
        "text": "hello"
    }
    result = _serialize_value(test_dict)
    
    assert result["numbers"] == ["1", "2", "3"]
    assert result["text"] == "hello"


@pytest.mark.unit
def test_serialize_value_with_object_having_dict():
    """
    Test that _serialize_value serializes objects with __dict__ attribute.
    """
    class TestObject:
        def __init__(self):
            self.attr1 = "value1"
            self.attr2 = 42
    
    obj = TestObject()
    result = _serialize_value(obj)
    
    assert isinstance(result, dict)
    assert result["attr1"] == "value1"
    assert result["attr2"] == "42"


@pytest.mark.unit
def test_serialize_value_with_nested_object():
    """
    Test that _serialize_value recursively serializes nested objects.
    """
    class InnerObject:
        def __init__(self):
            self.inner_attr = "inner_value"
    
    class OuterObject:
        def __init__(self):
            self.outer_attr = "outer_value"
            self.inner = InnerObject()
    
    obj = OuterObject()
    result = _serialize_value(obj)
    
    assert isinstance(result, dict)
    assert result["outer_attr"] == "outer_value"
    assert result["inner"]["inner_attr"] == "inner_value"


@pytest.mark.unit
def test_serialize_value_with_integer():
    """
    Test that _serialize_value converts integers to string representation.
    """
    result = _serialize_value(12345)
    assert result == "12345"


@pytest.mark.unit
def test_serialize_value_with_float():
    """
    Test that _serialize_value converts floats to string representation.
    """
    result = _serialize_value(3.14159)
    assert result == "3.14159"


@pytest.mark.unit
def test_serialize_value_with_boolean():
    """
    Test that _serialize_value converts booleans to string representation.
    """
    result_true = _serialize_value(True)
    result_false = _serialize_value(False)
    
    assert result_true == "True"
    assert result_false == "False"


@pytest.mark.unit
def test_serialize_value_with_long_string_representation():
    """
    Test that _serialize_value truncates long string representations of objects.
    """
    class VerboseObject:
        __slots__ = []  # Prevent __dict__ so __str__ is used
        
        def __str__(self):
            return "x" * 1500
    
    obj = VerboseObject()
    result = _serialize_value(obj, max_length=1000)
    
    assert result.startswith("x" * 1000)
    assert "... (truncated)" in result


@pytest.mark.unit
def test_serialize_value_respects_max_length_for_strings():
    """
    Test that _serialize_value respects custom max_length parameter for strings.
    """
    test_string = "a" * 500
    result = _serialize_value(test_string, max_length=100)
    
    assert result.startswith("a" * 100)
    assert "... (truncated, total length: 500)" in result


@pytest.mark.unit
def test_serialize_value_respects_max_length_for_list_items():
    """
    Test that _serialize_value distributes max_length across list items.
    """
    long_string = "x" * 500
    test_list = [long_string, long_string]
    result = _serialize_value(test_list, max_length=100)
    
    # Each item should get max_length // len(list) = 50
    assert isinstance(result, list)
    assert all("truncated" in item for item in result)


@pytest.mark.unit
def test_serialize_value_respects_max_length_for_dict_values():
    """
    Test that _serialize_value distributes max_length across dict values.
    """
    long_string = "x" * 500
    test_dict = {"key1": long_string, "key2": long_string}
    result = _serialize_value(test_dict, max_length=100)
    
    # Each value should get max_length // len(dict) = 50
    assert isinstance(result, dict)
    assert all("truncated" in v for v in result.values())


@pytest.mark.unit
def test_serialize_value_with_large_list_and_max_length():
    """
    Test that _serialize_value creates proper sample for large lists with max_length constraint.
    """
    test_list = ["a" * 100 for _ in range(15)]
    result = _serialize_value(test_list, max_length=1000)
    
    assert isinstance(result, dict)
    assert result["type"] == "list"
    assert result["count"] == 15
    assert len(result["sample"]) == 3


@pytest.mark.unit
def test_serialize_value_with_complex_nested_structure():
    """
    Test that _serialize_value handles complex nested structures correctly.
    """
    complex_structure = {
        "users": [
            {"name": "Alice", "age": 30, "tags": ["admin", "active"]},
            {"name": "Bob", "age": 25, "tags": ["user"]}
        ],
        "metadata": {
            "count": 2,
            "filters": ["active", "verified"]
        }
    }
    result = _serialize_value(complex_structure)
    
    assert isinstance(result, dict)
    assert result["users"][0]["name"] == "Alice"
    assert result["users"][1]["tags"] == ["user"]
    assert result["metadata"]["count"] == "2"
    assert result["metadata"]["filters"] == ["active", "verified"]


@pytest.mark.unit
def test_serialize_value_with_tuple():
    """
    Test that _serialize_value converts tuples to string representation.
    """
    test_tuple = (1, 2, 3)
    result = _serialize_value(test_tuple)
    
    assert result == "(1, 2, 3)"


@pytest.mark.unit
def test_serialize_value_with_set():
    """
    Test that _serialize_value converts sets to string representation.
    """
    test_set = {1, 2, 3}
    result = _serialize_value(test_set)
    
    # Sets are unordered, so just check it's a string containing the elements
    assert isinstance(result, str)
    assert "1" in result
    assert "2" in result
    assert "3" in result


@pytest.mark.unit
def test_serialize_value_with_custom_class_no_dict():
    """
    Test that _serialize_value handles custom classes without __dict__.
    """
    class CustomClass:
        __slots__ = ['value']
        
        def __init__(self):
            self.value = 42
        
        def __str__(self):
            return f"CustomClass(value={self.value})"
    
    obj = CustomClass()
    result = _serialize_value(obj)
    
    assert isinstance(result, str)
    assert "CustomClass" in result
    assert "42" in result



import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from unittest.mock import MagicMock

from artifex.models.classification.binary_classification.spam_detection import SpamDetection
from artifex.core import ParsedModelInstructions
from artifex.config import config


@pytest.fixture
def mock_dependencies(mocker: MockerFixture) -> None:
    """
    Fixture to mock external dependencies for SpamDetection.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_model = MagicMock()
    mock_model.config.id2label = {0: "not_spam", 1: "spam"}
    
    mocker.patch(
        'artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained',
        return_value=mock_model
    )
    mocker.patch(
        'artifex.models.classification.classification_model.AutoTokenizer.from_pretrained',
        return_value=MagicMock()
    )


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def spam_detection(mock_dependencies: None, mock_synthex: Synthex) -> SpamDetection:
    """
    Fixture to create a SpamDetection instance for testing.
    
    Args:
        mock_dependencies (None): Fixture that mocks external dependencies.
        mock_synthex (Synthex): A mocked Synthex instance.
    
    Returns:
        SpamDetection: A SpamDetection instance.
    """
    
    return SpamDetection(synthex=mock_synthex)


@pytest.mark.unit
def test_parse_user_instructions_returns_parsed_model_instructions(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions returns a ParsedModelInstructions instance.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing links"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert isinstance(result, ParsedModelInstructions)


@pytest.mark.unit
def test_parse_user_instructions_sets_user_instructions_field(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions correctly sets the user_instructions field.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing links", "malware"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_sets_language_field(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions correctly sets the language field.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing links"]
    language = "spanish"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == language


@pytest.mark.unit
def test_parse_user_instructions_domain_is_none(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions sets domain field to None.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing links"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.domain is None


@pytest.mark.unit
def test_parse_user_instructions_with_single_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with a single spam content item.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["lottery scams"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["lottery scams"]
    assert result.language == "english"
    assert result.domain is None


@pytest.mark.unit
def test_parse_user_instructions_with_multiple_spam_contents(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with multiple spam content items.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing", "malware", "scams", "fraud"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["phishing", "malware", "scams", "fraud"]
    assert len(result.user_instructions) == 4


@pytest.mark.unit
def test_parse_user_instructions_with_empty_list(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with an empty list.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = []
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == []
    assert result.language == "english"


@pytest.mark.unit
def test_parse_user_instructions_with_unicode_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles unicode characters in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["诈骗", "フィッシング", "мошенничество"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["诈骗", "フィッシング", "мошенничество"]


@pytest.mark.unit
def test_parse_user_instructions_with_french_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with French language.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["hameçonnage"]
    language = "french"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "french"
    assert result.user_instructions == ["hameçonnage"]


@pytest.mark.unit
def test_parse_user_instructions_with_german_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with German language.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["Phishing-Versuche"]
    language = "german"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "german"


@pytest.mark.unit
def test_parse_user_instructions_with_italian_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with Italian language.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["truffe online"]
    language = "italian"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "italian"


@pytest.mark.unit
def test_parse_user_instructions_with_japanese_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with Japanese language.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["スパムメール"]
    language = "japanese"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "japanese"


@pytest.mark.unit
def test_parse_user_instructions_with_chinese_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with Chinese language.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["垃圾邮件"]
    language = "chinese"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "chinese"


@pytest.mark.unit
def test_parse_user_instructions_with_russian_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions works with Russian language.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["спам"]
    language = "russian"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "russian"


@pytest.mark.unit
def test_parse_user_instructions_with_special_characters(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles special characters in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["click here!!!", "buy now $$$", "FREE!!!"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["click here!!!", "buy now $$$", "FREE!!!"]


@pytest.mark.unit
def test_parse_user_instructions_with_long_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles long spam content descriptions.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    long_content = (
        "Emails pretending to be from legitimate financial institutions requesting "
        "sensitive personal information such as passwords, credit card numbers, or "
        "social security numbers"
    )
    user_instructions = [long_content]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == [long_content]


@pytest.mark.unit
def test_parse_user_instructions_with_quotes_in_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles quotes in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["Messages saying 'you won!'", 'Emails with "urgent action required"']
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["Messages saying 'you won!'", 'Emails with "urgent action required"']


@pytest.mark.unit
def test_parse_user_instructions_with_newlines_in_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles newlines in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["multi\nline\nspam"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["multi\nline\nspam"]


@pytest.mark.unit
def test_parse_user_instructions_with_tabs_in_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles tabs in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["content\twith\ttabs"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["content\twith\ttabs"]


@pytest.mark.unit
def test_parse_user_instructions_with_mixed_case_language(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions preserves language case.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing"]
    language = "English"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.language == "English"


@pytest.mark.unit
def test_parse_user_instructions_with_numeric_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles numeric content in spam descriptions.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["account number 123456", "win $1,000,000"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["account number 123456", "win $1,000,000"]


@pytest.mark.unit
def test_parse_user_instructions_with_very_long_list(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles a very long list of spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    spam_list = [f"spam_type_{i}" for i in range(50)]
    language = "english"
    
    result = spam_detection._parse_user_instructions(spam_list, language)
    
    assert result.user_instructions == spam_list
    assert len(result.user_instructions) == 50


@pytest.mark.unit
def test_parse_user_instructions_with_html_in_spam_content(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles HTML tags in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["<a href='phishing.com'>Click here</a>", "<b>URGENT</b>"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["<a href='phishing.com'>Click here</a>", "<b>URGENT</b>"]


@pytest.mark.unit
def test_parse_user_instructions_with_backslashes(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles backslashes in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["path\\to\\malware", "c:\\windows\\system32"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["path\\to\\malware", "c:\\windows\\system32"]


@pytest.mark.unit
def test_parse_user_instructions_with_emojis(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles emojis in spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["🎉 You won! 🎉", "💰 Click here 💰"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["🎉 You won! 🎉", "💰 Click here 💰"]


@pytest.mark.unit
def test_parse_user_instructions_with_whitespace_only(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles whitespace-only spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["   ", "\t\t", "\n\n"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["   ", "\t\t", "\n\n"]


@pytest.mark.unit
def test_parse_user_instructions_with_empty_strings(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions handles empty strings in the list.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["", "phishing", ""]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["", "phishing", ""]


@pytest.mark.unit
def test_parse_user_instructions_preserves_order(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions preserves the order of spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["third", "first", "second"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["third", "first", "second"]


@pytest.mark.unit
def test_parse_user_instructions_with_duplicates(
    spam_detection: SpamDetection
) -> None:
    """
    Test that _parse_user_instructions preserves duplicate spam content.
    
    Args:
        spam_detection (SpamDetection): The SpamDetection instance.
    """
    
    user_instructions = ["phishing", "malware", "phishing"]
    language = "english"
    
    result = spam_detection._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == ["phishing", "malware", "phishing"]
    assert len(result.user_instructions) == 3


import pytest

from artifex import Artifex
from artifex.core import ClassificationResponse


expected_labels = [
    "greeting", "farewell", "thank_you", "affirmation", "negation", "small_talk",
    "bot_capabilities", "feedback_positive", "feedback_negative", "clarification",
    "suggestion", "language_change"
]

@pytest.mark.integration
def test__call__single_input_success(
    artifex: Artifex
):
    """
    Test the `__call__` method of the `IntentClassifier` class when a single input is 
    provided. Ensure that:
    - It returns a list of ClassificationResponse objects.
    - The output labels are among the expected intent labels.
    Args:
        artifex (Artifex): The Artifex instance to be used for testing.
    """
    
    out = artifex.intent_classifier("test input", device=-1, disable_logging=True)
    assert isinstance(out, list)
    assert all(isinstance(resp, ClassificationResponse) for resp in out)
    assert all(resp.label in expected_labels for resp in out)
    
@pytest.mark.integration
def test__call__multiple_inputs_success(
    artifex: Artifex
):
    """
    Test the `__call__` method of the `IntentClassifier` class when multiple inputs are 
    provided. Ensure that:
    - It returns a list of ClassificationResponse objects.
    - The output labels are among the expected intent labels.
    Args:
        artifex (Artifex): The Artifex instance to be used for testing.
    """
    
    out = artifex.intent_classifier(
        ["test input 1", "test input 2", "test input 3"], device=-1, disable_logging=True
    )
    assert isinstance(out, list)
    assert all(isinstance(resp, ClassificationResponse) for resp in out)
    assert all(resp.label in expected_labels for resp in out)


from pydantic import validate_call, ValidationError
from typing import Any, Callable, TypeVar
from functools import wraps
import inspect


T = TypeVar("T", bound=type)


def _should_skip_method(attr: Any, attr_name: str) -> bool:
    """
    Determines whether a class attribute should be skipped based on its name and signature.
    This function skips:
    - Dunder (double underscore) methods, except for '__call__'.
    - Attributes that are not callable.
    - Methods that only have 'self' as their parameter.
    Args:
        cls (T): The class containing the attribute.
        attr_name (str): The name of the attribute to check.
    Returns:
        bool: True if the attribute should be skipped, False otherwise.
    """
    
    # Skip dunder methods, except the __call__ method
    if attr_name.startswith("__") and attr_name != "__call__":
        return True

    if not callable(attr):
        return True

    # Get method signature and skip methods that only have 'self' as parameter or no parameters at all
    sig = inspect.signature(attr)
    params = list(sig.parameters.values())
    if len(params) == 0:
        return True
    if len(params) == 1 and params[0].name == "self":
        return True
    
    return False

def auto_validate_methods(cls: T) -> T:
    """
    A class decorator that combines Pydantic's `validate_call` for input validation
    and automatic handling of validation errors, raising a custom `ArtifexValidationError`.
    """
    
    from artifex.core import ValidationError as ArtifexValidationError

    for attr_name in dir(cls):
        # Use getattr_static to avoid triggering descriptors
        raw_attr = inspect.getattr_static(cls, attr_name)
        attr = getattr(cls, attr_name)

        is_static = isinstance(raw_attr, staticmethod)
        is_class = isinstance(raw_attr, classmethod)

        # Unwrap only if it's a staticmethod/classmethod object
        if is_static or is_class:
            func = raw_attr.__func__
        else:
            func = attr

        if _should_skip_method(func, attr_name):
            continue

        validated = validate_call(config={"arbitrary_types_allowed": True})(func)

        @wraps(func)
        def wrapper(*args: Any, __f: Callable[..., Any] = validated, **kwargs: Any) -> Any:
            try:
                return __f(*args, **kwargs)
            except ValidationError as e:
                raise ArtifexValidationError(f"Invalid input: {e}")

        # Re-wrap as staticmethod/classmethod if needed
        if is_static:
            setattr(cls, attr_name, staticmethod(wrapper))
        elif is_class:
            setattr(cls, attr_name, classmethod(wrapper))
        else:
            setattr(cls, attr_name, wrapper)

    return cls


import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from datasets import ClassLabel
import pandas as pd
import tempfile
import os

from artifex.models import ClassificationModel


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        Synthex: A mocked Synthex instance.
    """
    return mocker.MagicMock()


@pytest.fixture
def concrete_model(mock_synthex: Synthex, mocker: MockerFixture) -> ClassificationModel:
    """
    Fixture to create a concrete ClassificationModel instance for testing.
    Args:
        mock_synthex (Synthex): A mocked Synthex instance.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        ClassificationModel: A concrete implementation of ClassificationModel.
    """
    # Mock the transformers components
    mocker.patch(
        "transformers.AutoModelForSequenceClassification.from_pretrained",
        return_value=mocker.MagicMock()
    )
    mocker.patch(
        "transformers.AutoTokenizer.from_pretrained",
        return_value=mocker.MagicMock()
    )
    
    class ConcreteNClassClassificationModel(ClassificationModel):
        """Concrete implementation of ClassificationModel for testing purposes."""
        
        @property
        def _base_model_name(self) -> str:
            return "distilbert-base-uncased"
        
        @property
        def _system_data_gen_instr(self) -> list[str]:
            return ["system instruction 1", "system instruction 2"]
        
        def _get_data_gen_instr(self, user_instr: list[str]) -> list[str]:
            return user_instr
    
    model = ConcreteNClassClassificationModel(mock_synthex)
    # Set up labels for testing
    model._labels = ClassLabel(names=["positive", "negative", "neutral"])
    
    return model


@pytest.fixture
def temp_csv_file():
    """
    Fixture to create a temporary CSV file for testing.
    Returns:
        str: Path to the temporary CSV file.
    """
    fd, path = tempfile.mkstemp(suffix=".csv")
    os.close(fd)
    yield path
    # Cleanup
    if os.path.exists(path):
        os.remove(path)


@pytest.mark.unit
def test_cleanup_removes_rows_with_invalid_labels(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset removes rows with invalid labels.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    # Create test data with invalid label
    df = pd.DataFrame({
        "text": ["This is a valid text", "Another valid text"],
        "labels": ["positive", "invalid_label"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1
    assert result_df.iloc[0]["labels"] == 0  # "positive" converted to index


@pytest.mark.unit
def test_cleanup_removes_rows_with_short_text(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset removes rows with text shorter than 10 characters.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["Short", "This is a valid text string"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1
    assert result_df.iloc[0]["text"] == "This is a valid text string"


@pytest.mark.unit
def test_cleanup_removes_rows_with_empty_text(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset removes rows with empty text.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["", "This is a valid text string"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1
    assert result_df.iloc[0]["text"] == "This is a valid text string"


@pytest.mark.unit
def test_cleanup_converts_labels_to_indexes(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset converts string labels to indexes.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is positive text", "This is negative text", "This is neutral text"],
        "labels": ["positive", "negative", "neutral"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert result_df.iloc[0]["labels"] == 0  # positive
    assert result_df.iloc[1]["labels"] == 1  # negative
    assert result_df.iloc[2]["labels"] == 2  # neutral


@pytest.mark.unit
def test_cleanup_keeps_valid_rows(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset keeps all valid rows.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is positive text", "This is negative text", "This is neutral text"],
        "labels": ["positive", "negative", "neutral"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 3


@pytest.mark.unit
def test_cleanup_removes_whitespace_only_text(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset removes rows with whitespace-only text.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["     ", "This is a valid text string"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1


@pytest.mark.unit
def test_cleanup_handles_text_with_exactly_10_characters(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset keeps text with exactly 10 characters.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["1234567890", "This is longer text"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 2


@pytest.mark.unit
def test_cleanup_handles_text_with_9_characters(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset removes text with 9 characters.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["123456789", "This is longer text"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1
    assert result_df.iloc[0]["text"] == "This is longer text"


@pytest.mark.unit
def test_cleanup_handles_mixed_valid_invalid_rows(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset correctly filters mixed valid/invalid rows.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": [
            "This is valid text",
            "Short",
            "Another valid text here",
            "",
            "Valid neutral text"
        ],
        "labels": ["positive", "negative", "invalid", "neutral", "neutral"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 2  # Only first and last rows are valid


@pytest.mark.unit
def test_cleanup_preserves_column_order(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset preserves column order.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is valid text"],
        "labels": ["positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert list(result_df.columns) == ["text", "labels"]


@pytest.mark.unit
def test_cleanup_saves_to_same_file(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset saves to the same file path.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is valid text"],
        "labels": ["positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    original_mtime = os.path.getmtime(temp_csv_file)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    assert os.path.exists(temp_csv_file)
    # File should have been modified
    new_mtime = os.path.getmtime(temp_csv_file)
    assert new_mtime >= original_mtime


@pytest.mark.unit
def test_cleanup_handles_text_with_leading_whitespace(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset strips leading whitespace when checking length.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["   Short", "This is valid text"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1
    assert result_df.iloc[0]["text"] == "This is valid text"


@pytest.mark.unit
def test_cleanup_handles_text_with_trailing_whitespace(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset strips trailing whitespace when checking length.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["Short   ", "This is valid text"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1


@pytest.mark.unit
def test_cleanup_uses_last_column_for_labels(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset uses the last column as labels.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is valid text"],
        "extra_column": ["some data"],
        "labels": ["positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    # Label should be converted to index
    assert result_df.iloc[0, -1] == 0


@pytest.mark.unit
def test_cleanup_uses_first_column_for_text(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset uses the first column as text.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["Short", "This is valid text"],
        "extra_column": ["data1", "data2"],
        "labels": ["positive", "negative"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    # Only the row with long enough text in first column should remain
    assert len(result_df) == 1
    assert result_df.iloc[0, 0] == "This is valid text"


@pytest.mark.unit
def test_cleanup_with_all_invalid_rows(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset handles dataset with all invalid rows.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["Short", "", "   "],
        "labels": ["positive", "negative", "neutral"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 0


@pytest.mark.unit
def test_cleanup_with_multiple_label_types(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset correctly converts all label types.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": [
            "This is positive text",
            "This is negative text",
            "This is neutral text",
            "Another positive text"
        ],
        "labels": ["positive", "negative", "neutral", "positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert result_df.iloc[0]["labels"] == 0  # positive
    assert result_df.iloc[1]["labels"] == 1  # negative
    assert result_df.iloc[2]["labels"] == 2  # neutral
    assert result_df.iloc[3]["labels"] == 0  # positive again


@pytest.mark.unit
def test_cleanup_does_not_add_index_column(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset does not add an index column.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is valid text"],
        "labels": ["positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    # Should only have 2 columns: text and labels
    assert len(result_df.columns) == 2
    assert "Unnamed: 0" not in result_df.columns


@pytest.mark.unit
def test_cleanup_with_case_sensitive_labels(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset is case-sensitive for label matching.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is valid text", "Another valid text"],
        "labels": ["Positive", "positive"]  # Capital P should be invalid
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    # Only lowercase "positive" should remain
    assert len(result_df) == 1
    assert result_df.iloc[0]["labels"] == 0


@pytest.mark.unit
def test_cleanup_preserves_text_content(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset preserves text content exactly.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    original_text = "This is some valid text with special chars !@#$%"
    df = pd.DataFrame({
        "text": [original_text],
        "labels": ["positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert result_df.iloc[0]["text"] == original_text


@pytest.mark.unit
def test_cleanup_handles_unicode_text(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset handles unicode text correctly.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    df = pd.DataFrame({
        "text": ["This is valid 日本語 text"],
        "labels": ["positive"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    assert len(result_df) == 1
    assert "日本語" in result_df.iloc[0]["text"]


@pytest.mark.unit
def test_cleanup_with_valid_labels_property(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset uses the model"s _labels property correctly.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    # Model has labels: ["positive", "negative", "neutral"]
    df = pd.DataFrame({
        "text": ["Valid text one", "Valid text two", "Valid text three"],
        "labels": ["positive", "negative", "neutral"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    # All three labels are valid
    assert len(result_df) == 3


@pytest.mark.unit
def test_cleanup_filters_with_model_labels(
    concrete_model: ClassificationModel,
    temp_csv_file: str
):
    """
    Test that _post_process_synthetic_dataset only keeps labels that exist in model"s _labels.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        temp_csv_file (str): Path to temporary CSV file.
    """
    
    # Model has labels: ["positive", "negative", "neutral"]
    df = pd.DataFrame({
        "text": ["Valid text one", "Valid text two", "Valid text three"],
        "labels": ["positive", "unknown", "neutral"]
    })
    df.to_csv(temp_csv_file, index=False)
    
    concrete_model._post_process_synthetic_dataset(temp_csv_file)
    
    result_df = pd.read_csv(temp_csv_file)
    # "unknown" should be filtered out
    assert len(result_df) == 2


import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from unittest.mock import MagicMock

from artifex.models.classification.binary_classification.guardrail import Guardrail
from artifex.core import ParsedModelInstructions
from artifex.config import config


@pytest.fixture
def mock_dependencies(mocker: MockerFixture) -> None:
    """
    Fixture to mock external dependencies for Guardrail.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    """
    
    mock_model = MagicMock()
    mock_model.config.id2label = {0: "safe", 1: "unsafe"}
    
    mocker.patch(
        'artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained',
        return_value=mock_model
    )
    mocker.patch(
        'artifex.models.classification.classification_model.AutoTokenizer.from_pretrained',
        return_value=MagicMock()
    )


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def guardrail(mock_dependencies: None, mock_synthex: Synthex) -> Guardrail:
    """
    Fixture to create a Guardrail instance for testing.
    
    Args:
        mock_dependencies (None): Fixture that mocks external dependencies.
        mock_synthex (Synthex): A mocked Synthex instance.
    
    Returns:
        Guardrail: A Guardrail instance.
    """
    
    return Guardrail(synthex=mock_synthex)


@pytest.mark.unit
def test_parse_user_instructions_returns_parsed_model_instructions(
    guardrail: Guardrail
) -> None:
    """
    Test that _parse_user_instructions returns a ParsedModelInstructions instance.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert isinstance(result, ParsedModelInstructions)


@pytest.mark.unit
def test_parse_user_instructions_sets_user_instructions_field(
    guardrail: Guardrail
) -> None:
    """
    Test that _parse_user_instructions correctly sets the user_instructions field.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech", "violence"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_sets_language_field(
    guardrail: Guardrail
) -> None:
    """
    Test that _parse_user_instructions correctly sets the language field.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "spanish"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.language == "spanish"


@pytest.mark.unit
def test_parse_user_instructions_domain_is_none(
    guardrail: Guardrail
) -> None:
    """
    Test that _parse_user_instructions sets domain to None (no domain for Guardrail).
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.domain is None


@pytest.mark.unit
def test_parse_user_instructions_with_single_unsafe_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with single unsafe content item.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert len(result.user_instructions) == 1
    assert result.user_instructions[0] == "hate speech"


@pytest.mark.unit
def test_parse_user_instructions_with_multiple_unsafe_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with multiple unsafe content items.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech", "violence", "profanity"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert len(result.user_instructions) == 3
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_empty_list(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with empty unsafe content list.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = []
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == []
    assert result.language == "english"


@pytest.mark.unit
def test_parse_user_instructions_preserves_order(
    guardrail: Guardrail
) -> None:
    """
    Test that _parse_user_instructions preserves the order of unsafe content items.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["first", "second", "third"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "first"
    assert result.user_instructions[1] == "second"
    assert result.user_instructions[2] == "third"


@pytest.mark.unit
def test_parse_user_instructions_with_different_languages(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with different language values.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    
    for language in ["english", "spanish", "french", "german", "chinese"]:
        result = guardrail._parse_user_instructions(user_instructions, language)
        assert result.language == language


@pytest.mark.unit
def test_parse_user_instructions_with_special_characters(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with special characters in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech!@#$%", "violence&*()"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_unicode_characters(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with unicode characters in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["discurso de odio 你好", "暴力"]
    language = "spanish"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions
    assert result.language == "spanish"


@pytest.mark.unit
def test_parse_user_instructions_with_unicode_language(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with unicode language parameter.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "中文"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.language == "中文"


@pytest.mark.unit
def test_parse_user_instructions_with_whitespace_in_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with whitespace in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["  hate speech  ", "violence   "]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    # Whitespace should be preserved
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_long_content_strings(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with long unsafe content strings.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    long_content = "This is a very long description of unsafe content that spans multiple sentences."
    user_instructions = [long_content]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == long_content


@pytest.mark.unit
def test_parse_user_instructions_with_many_items(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with many unsafe content items.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = [f"unsafe_content_{i}" for i in range(20)]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert len(result.user_instructions) == 20
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_newlines_in_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with newlines in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["content\nwith\nnewlines"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "content\nwith\nnewlines"


@pytest.mark.unit
def test_parse_user_instructions_with_quotes_in_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with quotes in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ['content with "double quotes"', "content with 'single quotes'"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_empty_strings_in_list(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with empty strings in unsafe content list.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["", "hate speech", ""]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_numeric_strings(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with numeric strings in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["123", "456.789"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions == user_instructions


@pytest.mark.unit
def test_parse_user_instructions_does_not_modify_input(
    guardrail: Guardrail
) -> None:
    """
    Test that _parse_user_instructions does not modify the input list.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech", "violence"]
    original_instructions = user_instructions.copy()
    language = "english"
    
    guardrail._parse_user_instructions(user_instructions, language)
    
    assert user_instructions == original_instructions


@pytest.mark.unit
def test_parse_user_instructions_with_mixed_case_language(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with mixed case language parameter.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "EnGLisH"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.language == "EnGLisH"


@pytest.mark.unit
def test_parse_user_instructions_with_complex_punctuation(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with complex punctuation in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["content (with [nested {punctuation}])"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "content (with [nested {punctuation}])"


@pytest.mark.unit
def test_parse_user_instructions_with_emoji_in_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with emoji characters in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech 😠💢"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "hate speech 😠💢"


@pytest.mark.unit
def test_parse_user_instructions_with_backslashes(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with backslashes in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["content\\with\\backslashes"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "content\\with\\backslashes"


@pytest.mark.unit
def test_parse_user_instructions_with_tabs_in_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with tab characters in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["content\twith\ttabs"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "content\twith\ttabs"


@pytest.mark.unit
def test_parse_user_instructions_consecutive_calls_produce_same_result(
    guardrail: Guardrail
) -> None:
    """
    Test that consecutive calls with same input produce same result.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech", "violence"]
    language = "english"
    
    result1 = guardrail._parse_user_instructions(user_instructions, language)
    result2 = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result1.user_instructions == result2.user_instructions
    assert result1.language == result2.language
    assert result1.domain == result2.domain


@pytest.mark.unit
def test_parse_user_instructions_with_semicolons_in_content(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with semicolons in unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["content; with; semicolons"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert result.user_instructions[0] == "content; with; semicolons"


@pytest.mark.unit
def test_parse_user_instructions_user_instructions_is_list(
    guardrail: Guardrail
) -> None:
    """
    Test that user_instructions field in result is a list.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech"]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert isinstance(result.user_instructions, list)


@pytest.mark.unit
def test_parse_user_instructions_all_fields_are_set(
    guardrail: Guardrail
) -> None:
    """
    Test that all fields of ParsedModelInstructions are properly set.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = ["hate speech", "violence"]
    language = "french"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert hasattr(result, 'user_instructions')
    assert hasattr(result, 'language')
    assert hasattr(result, 'domain')
    assert result.user_instructions is not None
    assert result.language is not None


@pytest.mark.unit
def test_parse_user_instructions_with_mixed_content_types(
    guardrail: Guardrail
) -> None:
    """
    Test _parse_user_instructions with mixed types of unsafe content.
    
    Args:
        guardrail (Guardrail): The Guardrail instance.
    """
    
    user_instructions = [
        "short",
        "This is a longer description of unsafe content.",
        "unicode 你好",
        "special!@#$%"
    ]
    language = "english"
    
    result = guardrail._parse_user_instructions(user_instructions, language)
    
    assert len(result.user_instructions) == 4
    assert result.user_instructions == user_instructions


from transformers import Trainer, TrainerState, TrainingArguments, TrainerCallback, TrainerControl
from transformers.trainer_utils import TrainOutput
from typing import Any, Dict
from rich.progress import Progress, BarColumn, TextColumn, TimeElapsedColumn, TaskID
from rich.console import Console

"""
Patches for HuggingFace classes to improve user experience.
"""

console = Console()

class SilentTrainer(Trainer):
    """
    A regular transformers.Trainer which prevents the tedious final training summary dictionary
    from being printed to the console (since as of now, there is no built-in way to disable it).
    """
    
    def train(self, *args: Any, **kwargs: Any) -> TrainOutput:
        import builtins
        orig_print = builtins.print

        def silent_print(*a: Any, **k: Any) -> None:
            # Only suppress the summary dictionary
            if (
                len(a) == 1
                and isinstance(a[0], dict)
                and "train_runtime" in a[0]
            ):
                return
            return orig_print(*a, **k)

        builtins.print = silent_print
        try:
            return super().train(*args, **kwargs)
        finally:
            builtins.print = orig_print
            
            
class RichProgressCallback(TrainerCallback):
    """
    A custom TrainerCallback that uses Rich to display a progress bar during training.
    """
    
    progress: Progress
    task: int

    def on_train_begin(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        Called at the beginning of training.
        """
        
        self.progress = Progress(
            TextColumn("[bold blue]{task.description}"),
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeElapsedColumn(),
            transient=True
        )
        self.task = self.progress.add_task("Training model...", total=state.max_steps)
        self.progress.start()

    def on_step_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        Called at the end of each training step.
        """
        
        self.progress.update(TaskID(self.task), completed=state.global_step)

    def on_train_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs: Dict[str, Any]
    ) -> None:
        """
        Called at the end of training.
        """
        
        self.progress.stop()
        console.print("[green]✔ Training model[/green]")


import pytest
from pytest_mock import MockerFixture
from datasets import ClassLabel
from typing import Any
from transformers.trainer_utils import TrainOutput
from transformers import AutoConfig

from artifex.models import NamedEntityRecognition
from artifex.core import ValidationError
from artifex.config import config


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Any:
    """
    Create a mock Synthex instance.    
    Args:
        mocker: pytest-mock fixture for creating mocks.        
    Returns:
        Mock Synthex instance.
    """
    return mocker.Mock()


@pytest.fixture
def ner_instance(mock_synthex: Any, mocker: MockerFixture) -> NamedEntityRecognition:
    """
    Create a NamedEntityRecognition instance with fully mocked dependencies.    
    Args:
        mock_synthex: Mocked Synthex instance.
        mocker: pytest-mock fixture for creating mocks.        
    Returns:
        NamedEntityRecognition instance with mocked components.
    """
    # Mock all external dependencies at module level
    mocker.patch("artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained")
    mocker.patch("artifex.models.named_entity_recognition.named_entity_recognition.AutoTokenizer.from_pretrained")
    
    # Mock config to avoid external dependencies
    mock_config = mocker.patch("artifex.models.named_entity_recognition.named_entity_recognition.config")
    mock_config.NER_HF_BASE_MODEL = "mock-model"
    mock_config.NER_TOKENIZER_MAX_LENGTH = 512
    mock_config.DEFAULT_SYNTHEX_DATAPOINT_NUM = 100
    mock_config.NER_TAGNAME_MAX_LENGTH = 50
    
    ner = NamedEntityRecognition(mock_synthex)
    
    return ner


@pytest.mark.unit
def test_train_validates_named_entity_names(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train validates named entity names using NERTagName.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    # Mock _train_pipeline to avoid actual training
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    # Mock AutoConfig
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    # Valid entity names should work
    named_entities = {
        "PERSON": "A person's name",
        "LOCATION": "A location"
    }
    
    # Should not raise
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain",
        output_path="/test"
    )


@pytest.mark.unit
def test_train_raises_validation_error_for_invalid_entity_names(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train raises ValidationError for invalid entity names.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    
    # Mock NERTagName to raise ValueError for invalid names
    mock_ner_tag_name = mocker.patch("artifex.models.named_entity_recognition.named_entity_recognition.NERTagName")
    mock_ner_tag_name.side_effect = ValueError("Invalid tag name")
    
    named_entities = {
        "INVALID TAG": "Has spaces"
    }
    
    with pytest.raises(ValidationError) as exc_info:
        ner_instance.train(
            named_entities=named_entities,
            domain="test domain"
        )
    
    assert "must be non-empty strings with no spaces" in str(exc_info.value.message)


@pytest.mark.unit
def test_train_creates_bio_labels_correctly(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train creates BIO labels (O, B-TAG, I-TAG) for each entity.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    named_entities = {
        "PERSON": "A person's name",
        "LOCATION": "A location"
    }
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    # Check that labels were set correctly
    expected_labels = ["O", "B-PERSON", "I-PERSON", "B-LOCATION", "I-LOCATION"]
    assert set(ner_instance._labels.names) == set(expected_labels)


@pytest.mark.unit
def test_train_includes_o_label(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train includes the 'O' (outside) label.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    named_entities = {"PERSON": "A person"}
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    assert "O" in ner_instance._labels.names


@pytest.mark.unit
def test_train_configures_model_with_correct_num_labels(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train configures the model with the correct number of labels.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mock_model = mocker.Mock()
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained",
        return_value=mock_model
    )
    
    named_entities = {
        "PERSON": "A person",
        "LOCATION": "A location"
    }
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    # Should have O, B-PERSON, I-PERSON, B-LOCATION, I-LOCATION = 5 labels
    assert mock_config.num_labels == 5


@pytest.mark.unit
def test_train_creates_id2label_mapping(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train creates correct id2label mapping in model config.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    named_entities = {"PERSON": "A person"}
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    # Verify id2label was set
    assert hasattr(mock_config, "id2label")
    assert isinstance(mock_config.id2label, dict)
    assert 0 in mock_config.id2label


@pytest.mark.unit
def test_train_creates_label2id_mapping(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train creates correct label2id mapping in model config.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    named_entities = {"PERSON": "A person"}
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    # Verify label2id was set
    assert hasattr(mock_config, "label2id")
    assert isinstance(mock_config.label2id, dict)
    assert "O" in mock_config.label2id


@pytest.mark.unit
def test_train_loads_model_with_updated_config(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train loads model from pretrained with updated config.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mock_from_pretrained = mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    named_entities = {"PERSON": "A person"}
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    # Verify from_pretrained was called with config
    call_kwargs = mock_from_pretrained.call_args[1]
    assert call_kwargs["config"] == mock_config
    assert call_kwargs["ignore_mismatched_sizes"] is True


@pytest.mark.unit
def test_train_calls_parse_user_instructions(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train calls _parse_user_instructions with NERInstructions.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mock_parse = mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["parsed_instruction"]
    )
    
    named_entities = {"PERSON": "A person"}
    domain = "medical records"
    
    ner_instance.train(
        named_entities=named_entities,
        domain=domain
    )
    
    # Verify _parse_user_instructions was called
    assert mock_parse.called
    call_args = mock_parse.call_args[0][0]
    assert call_args.named_entity_tags == named_entities
    assert call_args.domain == domain


@pytest.mark.unit
def test_train_calls_train_pipeline_with_correct_args(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train calls _train_pipeline with correct arguments.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction1", "instruction2"]
    )
    
    named_entities = {"PERSON": "A person"}
    output_path = "/test/output"
    num_samples = 200
    num_epochs = 5
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain",
        output_path=output_path,
        num_samples=num_samples,
        num_epochs=num_epochs
    )
    
    # Verify _train_pipeline was called with correct arguments
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["user_instructions"] == ["instruction1", "instruction2"]
    assert call_kwargs["output_path"] == output_path
    assert call_kwargs["num_samples"] == num_samples
    assert call_kwargs["num_epochs"] == num_epochs


@pytest.mark.unit
def test_train_passes_train_datapoint_examples(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train passes train_datapoint_examples to _train_pipeline.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    examples = [
        {"text": "John works", "labels": "John: PERSON"},
        {"text": "Paris is nice", "labels": "Paris: LOCATION"}
    ]
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test",
        train_datapoint_examples=examples
    )
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["train_datapoint_examples"] == examples


@pytest.mark.unit
def test_train_returns_train_output(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train returns the TrainOutput from _train_pipeline.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    expected_output = TrainOutput(
        global_step=150,
        training_loss=0.3,
        metrics={"accuracy": 0.95}
    )
    
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=expected_output
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    result = ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test"
    )
    
    assert result == expected_output


@pytest.mark.unit
def test_train_uses_default_num_samples(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train uses default num_samples from config when not provided.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    # Get the default value from config
    default_samples = config.DEFAULT_SYNTHEX_DATAPOINT_NUM
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test"
    )
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["num_samples"] == default_samples


@pytest.mark.unit
def test_train_uses_default_num_epochs(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train uses default num_epochs (3) when not provided.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test"
    )
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["num_epochs"] == 3


@pytest.mark.unit
def test_train_handles_multiple_entity_types(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train correctly handles multiple entity types.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    named_entities = {
        "PERSON": "A person's name",
        "LOCATION": "A geographic location",
        "ORGANIZATION": "A company or institution",
        "DATE": "A date or time"
    }
    
    ner_instance.train(
        named_entities=named_entities,
        domain="test domain"
    )
    
    # Should have 1 + 4*2 = 9 labels (O + 4 entities * 2 BIO tags each)
    assert len(ner_instance._labels.names) == 9
    assert "O" in ner_instance._labels.names
    assert "B-PERSON" in ner_instance._labels.names
    assert "I-PERSON" in ner_instance._labels.names
    assert "B-LOCATION" in ner_instance._labels.names
    assert "I-LOCATION" in ner_instance._labels.names
    assert "B-ORGANIZATION" in ner_instance._labels.names
    assert "I-ORGANIZATION" in ner_instance._labels.names
    assert "B-DATE" in ner_instance._labels.names
    assert "I-DATE" in ner_instance._labels.names


@pytest.mark.unit
def test_train_updates_model_instance_variable(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train updates the _model instance variable with new model.    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.        
    """
    mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    new_model = mocker.Mock()
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained",
        return_value=new_model
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test"
    )
    
    assert ner_instance._model == new_model
    
    
@pytest.mark.unit
def test_train_passes_device_to_train_pipeline(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train passes device parameter to _train_pipeline.
    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test",
        device=0
    )
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["device"] == 0


@pytest.mark.unit
def test_train_passes_device_minus_1_to_train_pipeline(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train passes device=-1 to _train_pipeline for CPU/MPS.
    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test",
        device=-1
    )
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["device"] == -1


@pytest.mark.unit
def test_train_passes_device_none_to_train_pipeline(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train passes device=None to _train_pipeline when not specified.
    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test"
    )
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["device"] is None


@pytest.mark.unit
def test_train_uses_default_device_when_not_provided(
    ner_instance: NamedEntityRecognition,
    mocker: MockerFixture
):
    """
    Test that train uses default device (None) when device parameter is not provided.
    
    Args:
        ner_instance: NamedEntityRecognition instance.
        mocker: pytest-mock fixture.
    """
    mock_train_pipeline = mocker.patch.object(
        ner_instance,
        "_train_pipeline",
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )
    
    mock_config = mocker.Mock(spec=AutoConfig)
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoConfig.from_pretrained",
        return_value=mock_config
    )
    
    mocker.patch(
        "artifex.models.named_entity_recognition.named_entity_recognition.AutoModelForTokenClassification.from_pretrained"
    )
    
    mocker.patch.object(
        ner_instance,
        "_parse_user_instructions",
        return_value=["instruction"]
    )
    
    # Call without device parameter
    ner_instance.train(
        named_entities={"PERSON": "A person"},
        domain="test",
        output_path="/test"
    )
    
    # Verify device=None is passed to _train_pipeline
    call_kwargs = mock_train_pipeline.call_args[1]
    assert "device" in call_kwargs
    assert call_kwargs["device"] is None


import pytest

from artifex import Artifex


@pytest.mark.integration
def test_train_success(
    artifex: Artifex,
    output_folder: str
):
    """
    Test the `train` method of the `ClassificationModel` class. Verify that:
    - The training process completes without errors.
    Args:
        artifex (Artifex): The Artifex instance to be used for testing.
    """
    
    classes = {
        "class_a": "Description for class A.",
        "class_b": "Description for class B.",
        "class_c": "Description for class C."
    }
    
    tc = artifex.text_classification
    
    tc.train(
        domain="test domain",
        classes=classes,
        num_samples=40,
        num_epochs=1,
        output_path=output_folder,
        device=-1,
        language="english",
        disable_logging=True
    )


import pytest
from pytest_mock import MockerFixture
from synthex import Synthex
from datasets import ClassLabel
from transformers.trainer_utils import TrainOutput

from artifex.models import ClassificationModel
from artifex.core import ValidationError
from artifex.config import config


@pytest.fixture
def mock_synthex(mocker: MockerFixture) -> Synthex:
    """
    Fixture to create a mock Synthex instance.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        Synthex: A mocked Synthex instance.
    """
    
    return mocker.MagicMock()


@pytest.fixture
def mock_auto_config(mocker: MockerFixture) -> MockerFixture:
    """
    Fixture to mock AutoConfig.from_pretrained.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        MockerFixture: Mocked AutoConfig.from_pretrained method.
    """
    
    mock_config = mocker.MagicMock()
    return mocker.patch(
        'artifex.models.classification.classification_model.AutoConfig.from_pretrained',
        return_value=mock_config
    )


@pytest.fixture
def mock_auto_model(mocker: MockerFixture) -> MockerFixture:
    """
    Fixture to mock AutoModelForSequenceClassification.from_pretrained.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        MockerFixture: Mocked AutoModelForSequenceClassification.from_pretrained method.
    """
    
    mock_model = mocker.MagicMock()
    return mocker.patch(
        'artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained',
        return_value=mock_model
    )


@pytest.fixture
def mock_train_pipeline(mocker: MockerFixture) -> MockerFixture:
    """
    Fixture to mock _train_pipeline method.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        MockerFixture: Mocked _train_pipeline method.
    """
    
    return mocker.patch.object(
        ClassificationModel,
        '_train_pipeline',
        return_value=TrainOutput(global_step=100, training_loss=0.5, metrics={})
    )


@pytest.fixture
def mock_parse_user_instructions(mocker: MockerFixture) -> MockerFixture:
    """
    Fixture to mock _parse_user_instructions method.
    Args:
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        MockerFixture: Mocked _parse_user_instructions method.
    """
    
    return mocker.patch.object(
        ClassificationModel,
        '_parse_user_instructions',
        return_value=["positive: Positive sentiment", "negative: Negative sentiment", "Reviews"]
    )


@pytest.fixture
def concrete_model(mock_synthex: Synthex, mocker: MockerFixture) -> ClassificationModel:
    """
    Fixture to create a concrete ClassificationModel instance for testing.
    Args:
        mock_synthex (Synthex): A mocked Synthex instance.
        mocker (MockerFixture): The pytest-mock fixture for mocking.
    Returns:
        ClassificationModel: A concrete implementation of ClassificationModel.
    """
    # Patch all external dependencies at the correct import path
    mocker.patch(
        "artifex.models.classification.classification_model.AutoModelForSequenceClassification.from_pretrained",
        return_value=mocker.MagicMock()
    )
    mocker.patch(
        "artifex.models.classification.classification_model.AutoTokenizer.from_pretrained",
        return_value=mocker.MagicMock()
    )
    mocker.patch(
        "artifex.models.classification.classification_model.ClassLabel",
        side_effect=lambda names: ClassLabel(names=names)
    )
    mocker.patch("artifex.models.base_model.BaseModel.__init__", return_value=None)

    class ConcreteClassificationModel(ClassificationModel):
        """Concrete implementation of ClassificationModel for testing purposes."""
        @property
        def _base_model_name(self) -> str:
            return "distilbert-base-uncased"
        @property
        def _system_data_gen_instr(self) -> list[str]:
            return ["system instruction 1", "system instruction 2"]
        def _get_data_gen_instr(self, user_instr: list[str]) -> list[str]:
            return user_instr

    return ConcreteClassificationModel(mock_synthex)


@pytest.mark.unit
def test_train_validates_class_names(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train validates class names.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive sentiment", "negative": "Negative sentiment"}
    
    result = concrete_model.train(
        domain="Reviews",
        classes=classes
    )
    
    assert isinstance(result, TrainOutput)


@pytest.mark.unit
def test_train_raises_error_for_class_name_with_spaces(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train raises ValidationError for class names with spaces.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"invalid class": "Invalid class name"}
    
    with pytest.raises(ValidationError) as exc_info:
        concrete_model.train(domain="Reviews", classes=classes)
    
    assert "no spaces" in str(exc_info.value.message)


@pytest.mark.unit
def test_train_raises_error_for_too_long_class_name(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train raises ValidationError for class names exceeding max length.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    long_name = "a" * (config.CLASSIFICATION_CLASS_NAME_MAX_LENGTH + 1)
    classes = {long_name: "Description"}
    
    with pytest.raises(ValidationError) as exc_info:
        concrete_model.train(domain="Reviews", classes=classes)
    
    assert "maximum length" in str(exc_info.value.message)


@pytest.mark.unit
def test_train_populates_labels_property(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train populates the _labels property.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive sentiment", "negative": "Negative sentiment"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(concrete_model._labels, ClassLabel)
    assert set(concrete_model._labels.names) == {"positive", "negative"}


@pytest.mark.unit
def test_train_calls_auto_config_from_pretrained(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train calls AutoConfig.from_pretrained with base model name.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive sentiment"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    mock_auto_config.assert_called_once_with("distilbert-base-uncased")


@pytest.mark.unit
def test_train_sets_num_labels_in_config(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train sets num_labels in model config.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive", "negative": "Negative", "neutral": "Neutral"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    model_config = mock_auto_config.return_value
    assert model_config.num_labels == 3


@pytest.mark.unit
def test_train_sets_id2label_in_config(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train sets id2label mapping in model config.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive", "negative": "Negative"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    model_config = mock_auto_config.return_value
    assert 0 in model_config.id2label
    assert 1 in model_config.id2label
    assert set(model_config.id2label.values()) == {"positive", "negative"}


@pytest.mark.unit
def test_train_sets_label2id_in_config(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train sets label2id mapping in model config.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive", "negative": "Negative"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    model_config = mock_auto_config.return_value
    assert "positive" in model_config.label2id
    assert "negative" in model_config.label2id
    assert set(model_config.label2id.values()) == {0, 1}


@pytest.mark.unit
def test_train_creates_model_with_correct_config(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train creates model with the correct config.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    call_kwargs = mock_auto_model.call_args[1]
    assert call_kwargs['config'] == mock_auto_config.return_value
    assert call_kwargs['ignore_mismatched_sizes'] is True


@pytest.mark.unit
def test_train_creates_model_with_base_model_name(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train creates model with base_model_name.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    mock_auto_model.assert_called_once()
    assert mock_auto_model.call_args[0][0] == "distilbert-base-uncased"


@pytest.mark.unit
def test_train_calls_parse_user_instructions(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train calls _parse_user_instructions.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive sentiment"}
    domain = "Reviews"
    
    concrete_model.train(domain=domain, classes=classes)
    
    mock_parse_user_instructions.assert_called_once()


@pytest.mark.unit
def test_train_calls_train_pipeline_with_user_instructions(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train calls _train_pipeline with parsed user instructions.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    parsed_instructions = mock_parse_user_instructions.return_value
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['user_instructions'] == parsed_instructions


@pytest.mark.unit
def test_train_calls_train_pipeline_with_output_path(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train calls _train_pipeline with output_path.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    output_path = "/path/to/output"
    
    concrete_model.train(domain="Reviews", classes=classes, output_path=output_path)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['output_path'] == output_path


@pytest.mark.unit
def test_train_calls_train_pipeline_with_num_samples(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train calls _train_pipeline with num_samples.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    num_samples = 1000
    
    concrete_model.train(domain="Reviews", classes=classes, num_samples=num_samples)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['num_samples'] == 1000


@pytest.mark.unit
def test_train_calls_train_pipeline_with_num_epochs(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train calls _train_pipeline with num_epochs.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    num_epochs = 5
    
    concrete_model.train(domain="Reviews", classes=classes, num_epochs=num_epochs)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['num_epochs'] == 5


@pytest.mark.unit
def test_train_uses_default_num_samples(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train uses default num_samples when not provided.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['num_samples'] == config.DEFAULT_SYNTHEX_DATAPOINT_NUM


@pytest.mark.unit
def test_train_uses_default_num_epochs(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train uses default num_epochs (3) when not provided.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['num_epochs'] == 3


@pytest.mark.unit
def test_train_returns_train_output(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train returns TrainOutput from _train_pipeline.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    result = concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(result, TrainOutput)
    assert result.global_step == 100
    assert result.training_loss == 0.5


@pytest.mark.unit
def test_train_with_single_class(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train works with a single class.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive sentiment"}
    
    result = concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(result, TrainOutput)
    assert len(concrete_model._labels.names) == 1


@pytest.mark.unit
def test_train_with_many_classes(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train works with many classes.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {f"class{i}": f"Description {i}" for i in range(10)}
    
    result = concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(result, TrainOutput)
    assert len(concrete_model._labels.names) == 10


@pytest.mark.unit
def test_train_assigns_model_to_instance(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train assigns the created model to _model.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    assert concrete_model._model == mock_auto_model.return_value


@pytest.mark.unit
def test_train_with_underscores_in_class_names(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train accepts class names with underscores.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"very_positive": "Very positive sentiment"}
    
    result = concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(result, TrainOutput)


@pytest.mark.unit
def test_train_with_hyphens_in_class_names(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train accepts class names with hyphens.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"very-positive": "Very positive sentiment"}
    
    result = concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(result, TrainOutput)


@pytest.mark.unit
def test_train_with_numeric_class_names(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train accepts numeric class names.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"class1": "First class", "class2": "Second class"}
    
    result = concrete_model.train(domain="Reviews", classes=classes)
    
    assert isinstance(result, TrainOutput)


@pytest.mark.unit
def test_train_preserves_class_order_in_labels(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train preserves class order in labels.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"first": "First", "second": "Second", "third": "Third"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    # Check that labels are present (order may vary due to dict iteration)
    assert set(concrete_model._labels.names) == {"first", "second", "third"}


@pytest.mark.unit
def test_train_with_all_parameters(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train works with all parameters specified.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive", "negative": "Negative"}
    
    result = concrete_model.train(
        domain="Movie reviews",
        classes=classes,
        output_path="/path/to/output",
        num_samples=2000,
        num_epochs=10
    )
    
    assert isinstance(result, TrainOutput)
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs['output_path'] == "/path/to/output"
    assert call_kwargs['num_samples'] == 2000
    assert call_kwargs['num_epochs'] == 10


@pytest.mark.unit
def test_train_raises_error_for_empty_class_name(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train raises ValidationError for empty class name.
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    Returns:
        None
    """
    classes = {"": "Empty class name"}
    
    with pytest.raises(ValidationError):
        concrete_model.train(domain="Reviews", classes=classes)
        
        
@pytest.mark.unit
def test_train_passes_device_none_to_train_pipeline(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train passes device=None to _train_pipeline when not specified.
    
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["device"] is None


@pytest.mark.unit
def test_train_passes_device_0_to_train_pipeline(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train passes device=0 to _train_pipeline for GPU.
    
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes, device=0)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["device"] == 0


@pytest.mark.unit
def test_train_passes_device_minus_1_to_train_pipeline(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train passes device=-1 to _train_pipeline for CPU/MPS.
    
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive"}
    
    concrete_model.train(domain="Reviews", classes=classes, device=-1)
    
    call_kwargs = mock_train_pipeline.call_args[1]
    assert call_kwargs["device"] == -1


@pytest.mark.unit
def test_train_uses_default_device_when_not_provided(
    concrete_model: ClassificationModel,
    mock_auto_config: MockerFixture,
    mock_auto_model: MockerFixture,
    mock_train_pipeline: MockerFixture,
    mock_parse_user_instructions: MockerFixture
):
    """
    Test that train uses default device (None) when device parameter is not provided.
    
    Args:
        concrete_model (ClassificationModel): The concrete ClassificationModel instance.
        mock_auto_config (MockerFixture): Mocked AutoConfig.
        mock_auto_model (MockerFixture): Mocked AutoModelForSequenceClassification.
        mock_train_pipeline (MockerFixture): Mocked _train_pipeline.
        mock_parse_user_instructions (MockerFixture): Mocked _parse_user_instructions.
    """

    classes = {"positive": "Positive", "negative": "Negative"}
    
    # Call without device parameter
    concrete_model.train(
        domain="Reviews",
        classes=classes,
        output_path="/test"
    )
    
    # Verify device=None is passed to _train_pipeline
    call_kwargs = mock_train_pipeline.call_args[1]
    assert "device" in call_kwargs
    assert call_kwargs["device"] is None